@@ -22,7 +22,6 @@ useful helper functions. Bounding objects include:
 """
 
 import warnings
-import math
 import numpy as np
 from numpy import linalg
 from numpy import cov as mle_cov
@@ -41,8 +40,6 @@ __all__ = [
     "_friends_leaveoneout_radius"
 ]
 
-SQRTEPS = math.sqrt(float(np.finfo(np.float64).eps))
-
 
 class UnitCube:
     """
@@ -119,10 +116,6 @@ class Ellipsoid:
         self.n = len(ctr)  # dimension
         self.ctr = np.asarray(ctr)  # center coordinates
         self.cov = np.asarray(cov)  # covariance matrix
-        if axes is None:
-            self.axes = lalg.cholesky(cov, lower=True)  # transformation axes
-        else:
-            self.axes = axes
 
         # The eigenvalues (l) of `a` are (a^-2, b^-2, ...) where
         # (a, b, ...) are the lengths of principle axes.
@@ -134,9 +127,14 @@ class Ellipsoid:
             # is a product of squares of eigen values
             self.logvol = logvol_prefactor(self.n) + 0.5 * np.log(l).sum()
         else:
-            raise ValueError("The input precision matrix defining the "
-                             "ellipsoid {0} is apparently singular with "
-                             "l={1} and v={2}.".format(self.cov, l, v))
+            raise ValueError(
+                "The input precision matrix defining the "
+                f"ellipsoid {self.cov} is apparently singular with "
+                f"l={l} and v={v}.")
+        if axes is None:
+            self.axes = v @ np.diag(self.axlens)
+        else:
+            self.axes = axes
         if am is None:
             self.am = v @ np.diag(1. / l) @ v.T
             # precision matrix (inverse of covariance)
@@ -186,7 +184,7 @@ class Ellipsoid:
             self.cov = v @ np.diag(l1) @ v.T
             self.am = v @ np.diag(1 / l1) @ v.T
             self.axlens *= fax
-            self.axes = lalg.cholesky(self.cov, lower=True)
+            self.axes = self.axes @ np.diag(fax)
             # I don't quite know how to scale axes, so I rerun cholesky
         self.logvol = logvol
 
@@ -252,7 +250,7 @@ class Ellipsoid:
         overlap between the ellipsoid and the unit cube."""
 
         samples = [self.sample(rstate=rstate) for i in range(ndraws)]
-        nin = sum([unitcheck(x) for x in samples])
+        nin = sum((unitcheck(x) for x in samples))
 
         return 1. * nin / ndraws
 
@@ -467,10 +465,17 @@ class MultiEllipsoid:
 
             # Check how many ellipsoids the point lies within
             delts = (x[None, :] - self.ctrs)
-            q = (np.einsum('ai,aij,aj->a', delts, self.ams, delts) < 1).sum()
-
-            assert q > 0  # Should never fail
+            ell_masks = np.einsum('ai,aij,aj->a', delts, self.ams, delts)
+            q = (ell_masks < 1).sum()
 
+            if q == 0:
+                # Should never be the case but may
+                # happen due to numerical inaccuracies
+                q = (ell_masks <= 1).sum()
+                if q == 0:
+                    max_mask = ell_masks.max()
+                    raise RuntimeError(
+                        f'Ellipsoid check failed q=0, {max_mask}')
             if return_q:
                 # If `q` is being returned, assume the user wants to
                 # explicitly apply the `1. / q` acceptance criterion to
@@ -479,7 +484,8 @@ class MultiEllipsoid:
             else:
                 # If `q` is not being returned, assume the user wants this
                 # done internally so we repeat the loop if needed
-                if q == 1 or rstate.uniform() < (1. / q):
+                # random is faster than uniform
+                if q == 1 or rstate.random() < (1. / q):
                     return x, idx
 
     def samples(self, nsamples, rstate=None):
@@ -510,14 +516,14 @@ class MultiEllipsoid:
         samples = [
             self.sample(rstate=rstate, return_q=True) for i in range(ndraws)
         ]
-        qsum = sum([q for (x, idx, q) in samples])
-        logvol = np.log(ndraws * 1. / qsum) + self.logvol_tot
+        qsum = sum((1. / q for (x, idx, q) in samples))
+        logvol = np.log(qsum / ndraws) + self.logvol_tot
 
         if return_overlap:
             # Estimate the fractional amount of overlap with the
             # unit cube using the same set of samples.
-            qin = sum([q * unitcheck(x) for (x, idx, q) in samples])
-            overlap = 1. * qin / qsum
+            qin = sum((1. / q * unitcheck(x) for (x, idx, q) in samples))
+            overlap = qin / qsum
             return logvol, overlap
         else:
             return logvol
@@ -639,7 +645,7 @@ class RadFriends:
 
         detsign, detln = linalg.slogdet(self.am)
         assert detsign > 0
-        self.logvol_ball = (logvol_prefactor(self.n) - 0.5 * detln)
+        self.logvol_ball = logvol_prefactor(self.n) - 0.5 * detln
         self.expand = 1.
         self.funit = 1
 
@@ -704,7 +710,8 @@ class RadFriends:
                 idx = rstate.integers(nctrs)
                 x = ctrs[idx] + dx
                 q = self.overlap(x, ctrs)
-            if q == 1 or return_q or rstate.uniform() < (1. / q):
+            # random is faster than uniform
+            if q == 1 or return_q or rstate.random() < (1. / q):
                 if return_q:
                     return x, q
                 else:
@@ -737,18 +744,19 @@ class RadFriends:
         estimated fractional overlap with the unit cube."""
 
         # Estimate volume using Monte Carlo integration.
-        samples = [
+        samples = ([
             self.sample(ctrs, rstate=rstate, return_q=True)
             for i in range(ndraws)
-        ]
-        qsum = sum([q for (x, q) in samples])
-        logvol = np.log(1. * ndraws / qsum * len(ctrs)) + self.logvol_ball
+        ])
+        qs = np.array([_[1] for _ in samples])
+        qsum = np.sum(1. / qs)
+        logvol = np.log(1. / ndraws * qsum * len(ctrs)) + self.logvol_ball
 
         if return_overlap:
             # Estimate the fractional amount of overlap with the
             # unit cube using the same set of samples.
-            qin = sum([q * unitcheck(x) for (x, q) in samples])
-            overlap = 1. * qin / qsum
+            qin = sum((1. / q * unitcheck(x) for (x, q) in samples))
+            overlap = qin / qsum
             return logvol, overlap
         else:
             return logvol
@@ -970,7 +978,8 @@ class SupFriends:
                 # Check how many cubes the point lies within, passing over
                 # the `idx`-th cube `x` was sampled from.
                 q = self.overlap(x, ctrs)
-            if q == 1 or return_q or rstate.uniform() < (1. / q):
+            # random() is faster than uniform()
+            if q == 1 or return_q or rstate.random() < (1. / q):
                 if return_q:
                     return x, q
                 else:
@@ -997,7 +1006,7 @@ class SupFriends:
                            ctrs,
                            ndraws=10000,
                            rstate=None,
-                           return_overlap=False):
+                           return_overlap=True):
         """Using `ndraws` Monte Carlo draws, estimate the log volume of the
         *union* of cubes. If `return_overlap=True`, also returns the
         estimated fractional overlap with the unit cube."""
@@ -1007,14 +1016,14 @@ class SupFriends:
             self.sample(ctrs, rstate=rstate, return_q=True)
             for i in range(ndraws)
         ]
-        qsum = sum([q for (x, q) in samples])
-        logvol = np.log(1. * ndraws / qsum * len(ctrs)) + self.logvol_cube
+        qsum = sum((1. / q for (x, q) in samples))
+        logvol = np.log(1. * qsum / ndraws * len(ctrs)) + self.logvol_cube
 
         if return_overlap:
             # Estimate the fractional overlap with the unit cube using
             # the same set of samples.
-            qin = sum([q * unitcheck(x) for (x, q) in samples])
-            overlap = 1. * qin / qsum
+            qin = sum((1. / q * unitcheck(x) for (x, q) in samples))
+            overlap = qin / qsum
             return logvol, overlap
         else:
             return logvol
@@ -1168,7 +1177,11 @@ def randsphere(n, rstate=None):
     """Draw a point uniformly within an `n`-dimensional unit sphere."""
 
     z = rstate.standard_normal(size=n)  # initial n-dim vector
-    xhat = z * (rstate.uniform()**(1. / n) / lalg.norm(z, check_finite=False))  # scale
+    # notice I use random () instead of uniform
+    # and standard_norm instead of normal as those are faster
+    # as this is a time-critical function
+    xhat = z * (rstate.random()**(1. / n) / lalg.norm(z, check_finite=False)
+                )  # scale
     return xhat
 
 
@@ -1178,7 +1191,8 @@ def rand_choice(pb, rstate):
     The pb must sum to 1
     """
     p1 = np.cumsum(pb)
-    xr = rstate.uniform()
+    # random is faster than uniform
+    xr = rstate.random()
     return min(np.searchsorted(p1, xr), len(pb) - 1)
 
 
@@ -1224,12 +1238,7 @@ def improve_covar_mat(covar0, ntries=100, max_condition_number=1e12):
                         # some eigen values are too small
                         failed = 1
                     else:
-                        # eigen values are all right
-                        # checking if cholesky works
-                        axes = lalg.cholesky(covar,
-                                             lower=True,
-                                             check_finite=False)
-                        # if we are here we are done
+                        axes = eigvec @ np.diag(eigval**.5)
                         break
             else:
                 # complete failure
@@ -1253,8 +1262,8 @@ def improve_covar_mat(covar0, ntries=100, max_condition_number=1e12):
         warnings.warn("Failed to guarantee the ellipsoid axes will be "
                       "non-singular. Defaulting to a sphere.")
         covar = np.eye(ndim)  # default to identity
-        am = lalg.pinvh(covar)
-        axes = lalg.cholesky(covar, lower=True)
+        am = covar.copy()
+        axes = covar.copy()
     else:
         # invert the matrix using eigen decomposition
         am = eigvec @ np.diag(1. / eigval) @ eigvec.T
