@@ -25,14 +25,15 @@ Includes:
 
 import math
 import copy
+import warnings
 import numpy as np
-
 from .sampler import Sampler
 from .bounding import (UnitCube, Ellipsoid, MultiEllipsoid, RadFriends,
                        SupFriends, rand_choice)
 from .sampling import (sample_unif, sample_rwalk, sample_slice, sample_rslice,
                        sample_hslice)
-from .utils import unitcheck, get_enlarge_bootstrap
+from .utils import (unitcheck, get_enlarge_bootstrap, save_sampler,
+                    restore_sampler)
 
 __all__ = [
     "UnitCubeSampler", "SingleEllipsoidSampler", "MultiEllipsoidSampler",
@@ -67,7 +68,8 @@ class SuperSampler(Sampler):
                  pool,
                  use_pool,
                  kwargs=None,
-                 ncdim=0):
+                 ncdim=0,
+                 blob=False):
         # Initialize sampler.
         super().__init__(loglikelihood,
                          prior_transform,
@@ -79,7 +81,8 @@ class SuperSampler(Sampler):
                          queue_size,
                          pool,
                          use_pool,
-                         ncdim=ncdim)
+                         ncdim=ncdim,
+                         blob=blob)
         # Initialize method to propose a new starting point.
         self._PROPOSE = {
             'unif': self.propose_unif,
@@ -131,11 +134,14 @@ class SuperSampler(Sampler):
         self.walks = max(2, self.kwargs.get('walks', 25))
         self.facc = self.kwargs.get('facc', 0.5)
         self.facc = min(1., max(1. / self.walks, self.facc))
+        self.rwalk_history = {'naccept': 0, 'nreject': 0}
 
         # Initialize slice parameters.
         self.slices = self.kwargs.get('slices', 5)
         self.fmove = self.kwargs.get('fmove', 0.9)
         self.max_move = self.kwargs.get('max_move', 100)
+        self.slice_history = {'ncontract': 0, 'nexpand': 0}
+        self.hslice_history = {'nmove': 0, 'nreflect': 0, 'ncontract': 0}
 
     def propose_unif(self, *args):
         pass
@@ -143,11 +149,11 @@ class SuperSampler(Sampler):
     def propose_live(self, *args):
         pass
 
-    def update_unif(self, blob):
+    def update_unif(self, blob, update=True):
         """Filler function."""
         pass
 
-    def update_rwalk(self, blob):
+    def update_rwalk(self, blob, update=True):
         """Update the random walk proposal scale based on the current
         number of accepted/rejected steps.
         For rwalk the scale is important because it
@@ -155,9 +161,16 @@ class SuperSampler(Sampler):
         I.e. if scale is too large, the proposal efficiency will be very low
         so it's likely that we'll only do one random walk step at the time,
         thus producing very correlated chain.
+        The keyword update determines if we are just accumulating the number
+        of steps or actually adjusting the scale
         """
         self.scale = blob['scale']
-        accept, reject = blob['accept'], blob['reject']
+        hist = self.rwalk_history
+        hist['naccept'] += blob['accept']
+        hist['nreject'] += blob['reject']
+        if not update:
+            return
+        accept, reject = hist['naccept'], hist['nreject']
         facc = (1. * accept) / (accept + reject)
         # Here we are now trying to solve the Eqn
         # f0 = F(s) where F is the function
@@ -172,19 +185,31 @@ class SuperSampler(Sampler):
         # See also Robbins-Munro recursion which we don't follow
         # here because our coefficients a_k do not obey \sum a_k^2 = \infty
         self.scale *= math.exp((facc - self.facc) / self.ncdim / self.facc)
+        hist['naccept'] = 0
+        hist['nreject'] = 0
 
-    def update_slice(self, blob):
+    def update_slice(self, blob, update=True):
         """Update the slice proposal scale based on the relative
         size of the slices compared to our initial guess.
         For slice sampling the scale is only 'advisory' in the sense that
         the right scale will just speed up sampling as we'll have to expand
         or contract less. It won't affect the quality of the samples much.
+        The keyword update determines if we are just accumulating the number
+        of steps or actually adjusting the scale
         """
         # see https://www.ncbi.nlm.nih.gov/pmc/articles/PMC4063214/
         # also 2002.06212
         # https://www.tandfonline.com/doi/full/10.1080/10618600.2013.791193
         # and https://github.com/joshspeagle/dynesty/issues/260
-        nexpand, ncontract = max(blob['nexpand'], 1), blob['ncontract']
+        hist = self.slice_history
+
+        hist['nexpand'] += blob['nexpand']
+        hist['ncontract'] += blob['ncontract']
+        if blob['expansion_warning_set']:
+            self.kwargs['slice_doubling'] = True
+        if not update:
+            return
+        nexpand, ncontract = max(hist['nexpand'], 1), hist['ncontract']
         mult = (nexpand * 2. / (nexpand + ncontract))
         # avoid drastic updates to the scale factor limiting to factor
         # of two
@@ -192,22 +217,66 @@ class SuperSampler(Sampler):
         # Remember I can't apply the rule that scale < cube diagonal
         # because scale is multiplied by axes
         self.scale = self.scale * mult
+        hist['nexpand'] = 0
+        hist['ncontract'] = 0
 
-    def update_hslice(self, blob):
+    def update_hslice(self, blob, update=True):
         """Update the Hamiltonian slice proposal scale based
-        on the relative amount of time spent moving vs reflecting."""
-
-        nmove, nreflect = blob['nmove'], blob['nreflect']
-        ncontract = blob.get('ncontract', 0)
+        on the relative amount of time spent moving vs reflecting.
+        The keyword update determines if we are just accumulating the number
+        of steps or actually adjusting the scale
+        """
+        hist = self.hslice_history
+        hist['nmove'] += blob['nmove']
+        hist['nreflect'] += blob['nreflect']
+        hist['ncontract'] += blob.get('ncontract', 0)
+        if not update:
+            return
+        nmove, nreflect = hist['nmove'], hist['nreflect']
+        ncontract = hist['ncontract']
         fmove = (1. * nmove) / (nmove + nreflect + ncontract + 2)
         norm = max(self.fmove, 1. - self.fmove)
         self.scale *= math.exp((fmove - self.fmove) / norm)
+        hist['nmove'] = 0
+        hist['nreflect'] = 0
+        hist['ncontract'] = 0
 
-    def update_user(self, blob):
+    def update_user(self, blob, update=True):
         """Update the scale based on the user-defined update function."""
 
         if callable(self.custom_update):
-            self.scale = self.custom_update(blob, self.scale)
+            self.scale = self.custom_update(blob, self.scale, update=update)
+
+    def save(self, fname):
+        """
+        Save the state of the dynamic sampler in a file
+
+        Parameters
+        ----------
+        fname: string
+            Filename of the save file.
+
+        """
+        save_sampler(self, fname)
+
+    @staticmethod
+    def restore(fname, pool=None):
+        """
+        Restore the dynamic sampler from a file.
+        It is assumed that the file was created using .save() method
+        of DynamicNestedSampler or as a result of checkpointing during
+        run_nested()
+
+        Parameters
+        ----------
+        fname: string
+            Filename of the save file.
+        pool: object(optional)
+            The multiprocessing pool-like object that supports map()
+            calls that will be used in the restored object.
+
+        """
+        return restore_sampler(fname, pool=pool)
 
 
 class UnitCubeSampler(SuperSampler):
@@ -278,7 +347,8 @@ class UnitCubeSampler(SuperSampler):
                  pool,
                  use_pool,
                  kwargs=None,
-                 ncdim=0):
+                 ncdim=0,
+                 blob=False):
 
         # Initialize sampler.
         super().__init__(loglikelihood,
@@ -293,6 +363,7 @@ class UnitCubeSampler(SuperSampler):
                          pool,
                          use_pool,
                          ncdim=ncdim,
+                         blob=blob,
                          kwargs=kwargs or {})
 
         self.unitcube = UnitCube(self.ncdim)
@@ -399,6 +470,7 @@ class SingleEllipsoidSampler(SuperSampler):
                  pool,
                  use_pool,
                  kwargs=None,
+                 blob=False,
                  ncdim=0):
 
         # Initialize sampler.
@@ -414,6 +486,7 @@ class SingleEllipsoidSampler(SuperSampler):
                          pool,
                          use_pool,
                          ncdim=ncdim,
+                         blob=blob,
                          kwargs=kwargs or {})
 
         self.ell = Ellipsoid(np.zeros(self.ncdim), np.identity(self.ncdim))
@@ -443,13 +516,25 @@ class SingleEllipsoidSampler(SuperSampler):
         """Propose a new live point by sampling *uniformly*
         within the ellipsoid."""
 
+        threshold_warning = 10000
+        if self.ncdim != self.npdim and self.nonbounded is not None:
+            nonb = self.nonbounded[:self.ncdim]
+        else:
+            nonb = self.nonbounded
+        niter = 0
         while True:
             # Sample a point from the ellipsoid.
             u = self.ell.sample(rstate=self.rstate)
-
+            niter += 1
             # Check if `u` is within the unit cube.
-            if unitcheck(u, self.nonbounded):
+            if unitcheck(u, nonb):
                 break  # if it is, we're done!
+
+        if niter > threshold_warning:
+            with warnings.catch_warnings():
+                warnings.filterwarnings("once")
+                warnings.warn("Ellipsoid sampling is extremely inefficient")
+
         if self.npdim != self.ncdim:
             u = np.concatenate(
                 [u, self.rstate.uniform(0, 1, self.npdim - self.ncdim)])
@@ -545,6 +630,7 @@ class MultiEllipsoidSampler(SuperSampler):
                  pool,
                  use_pool,
                  kwargs=None,
+                 blob=False,
                  ncdim=0):
         # Initialize sampler.
         super().__init__(loglikelihood,
@@ -559,6 +645,7 @@ class MultiEllipsoidSampler(SuperSampler):
                          pool,
                          use_pool,
                          ncdim=ncdim,
+                         blob=blob,
                          kwargs=kwargs or {})
 
         self.mell = MultiEllipsoid(ctrs=[np.zeros(self.ncdim)],
@@ -589,18 +676,28 @@ class MultiEllipsoidSampler(SuperSampler):
         """Propose a new live point by sampling *uniformly* within
         the union of ellipsoids."""
 
+        threshold_warning = 10000
+
         if self.ncdim != self.npdim and self.nonbounded is not None:
             nonb = self.nonbounded[:self.ncdim]
         else:
             nonb = self.nonbounded
+
+        niter = 0
+
         while True:
             # Sample a point from the union of ellipsoids.
             # Returns the point `u`, ellipsoid index `idx`, and number of
             # overlapping ellipsoids `q` at position `u`.
             u, idx = self.mell.sample(rstate=self.rstate)
+            niter += 1
             # Check if the point is within the unit cube.
             if unitcheck(u, nonb):
                 break  # if successful, we're done!
+        if niter > threshold_warning:
+            with warnings.catch_warnings():
+                warnings.filterwarnings("once")
+                warnings.warn("Ellipsoid sampling is extremely inefficient")
         if self.ncdim != self.npdim:
             u = np.concatenate(
                 [u, self.rstate.uniform(0, 1, self.npdim - self.ncdim)])
@@ -721,6 +818,7 @@ class RadFriendsSampler(SuperSampler):
                  pool,
                  use_pool,
                  kwargs=None,
+                 blob=False,
                  ncdim=0):
 
         # Initialize sampler.
@@ -736,6 +834,7 @@ class RadFriendsSampler(SuperSampler):
                          pool,
                          use_pool,
                          ncdim=ncdim,
+                         blob=blob,
                          kwargs=kwargs or {})
 
         self.radfriends = RadFriends(self.ncdim)
@@ -871,6 +970,7 @@ class SupFriendsSampler(SuperSampler):
                  pool,
                  use_pool,
                  kwargs=None,
+                 blob=False,
                  ncdim=0):
 
         # Initialize sampler.
@@ -886,6 +986,7 @@ class SupFriendsSampler(SuperSampler):
                          pool,
                          use_pool,
                          ncdim=ncdim,
+                         blob=blob,
                          kwargs=kwargs or {})
 
         self.supfriends = SupFriends(self.ncdim)
