@@ -9,11 +9,16 @@ import sys
 import warnings
 import math
 import copy
+import time
+import os
+import shutil
 from collections import namedtuple
 from functools import partial
+import pickle as pickle_module
+# To allow replacing of the pickler
 import numpy as np
 from scipy.special import logsumexp
-
+from ._version import __version__ as DYNESTY_VERSION
 try:
     import tqdm
 except ImportError:
@@ -24,12 +29,11 @@ try:
 except ImportError:
     h5py = None
 
-from .results import Results, print_fn, results_substitute
-
 __all__ = [
     "unitcheck", "resample_equal", "mean_and_cov", "quantile", "jitter_run",
     "resample_run", "reweight_run", "unravel_run", "merge_runs", "kld_error",
-    "_merge_two", "_get_nsamps_samples_n", "get_enlarge_bootstrap"
+    "get_enlarge_bootstrap", "LoglOutput", "LogLikelihood", "RunRecord",
+    "DelayTimer"
 ]
 
 SQRTEPS = math.sqrt(float(np.finfo(np.float64).eps))
@@ -37,17 +41,81 @@ SQRTEPS = math.sqrt(float(np.finfo(np.float64).eps))
 IteratorResult = namedtuple('IteratorResult', [
     'worst', 'ustar', 'vstar', 'loglstar', 'logvol', 'logwt', 'logz',
     'logzvar', 'h', 'nc', 'worst_it', 'boundidx', 'bounditer', 'eff',
-    'delta_logz'
+    'delta_logz', 'blob'
 ])
 
-IteratorResultShort = namedtuple('IteratorResult', [
+IteratorResultShort = namedtuple('IteratorResultShort', [
     'worst', 'ustar', 'vstar', 'loglstar', 'nc', 'worst_it', 'boundidx',
-    'bounditer', 'eff'
+    'bounditer', 'eff', 'blob'
 ])
 
 
+class LoglOutput:
+    """
+    Class that encapsulates the output of the likelihood function.
+    The reason we need this wrapper is to preserve the blob associated with
+    the likelihood function.
+
+    """
+
+    def __init__(self, v, blob_flag):
+        """
+        Initialize the object
+        
+        Parameters
+        ----------
+        v: float or tuple
+            if blob_flag is true v have to be a tuple of logl and blob
+            if it is False v is just logl
+        blob_flag: boolean
+            flag to mark whether the v has a blob or not
+        """
+        if blob_flag:
+            self.val = v[0]
+            self.blob = v[1]
+        else:
+            self.val = v
+
+    def __lt__(self, v1):
+        """
+        Comparison override, we just use .val attribute in the comparison
+        """
+        return float(self) < float(v1)
+
+    def __gt__(self, v1):
+        """
+        Comparison override, we just use .val attribute in the comparison
+        """
+        return float(self) > float(v1)
+
+    def __le__(self, v1):
+        """
+        Comparison override, we just use .val attribute in the comparison
+        """
+        return float(self) <= float(v1)
+
+    def __ge__(self, v1):
+        """
+        Comparison override, we just use .val attribute in the comparison
+        """
+        return float(self) >= float(v1)
+
+    def __eq__(self, v1):
+        """
+        Comparison override, we just use .val attribute in the comparison
+        """
+        return float(self) == float(v1)
+
+    def __float__(self):
+        """
+        Comparison override, we just use .val attribute in the comparison
+        """
+        return float(self.val)
+
+
 class LogLikelihood:
-    """ Class that calls the likelihood function (using a pool if provided)
+    """
+    Class that calls the likelihood function (using a pool if provided)
     Also if requested it saves the history of evaluations
     """
 
@@ -56,10 +124,12 @@ class LogLikelihood:
                  ndim,
                  pool=None,
                  save=False,
-                 history_filename=None):
+                 history_filename=None,
+                 blob=False):
         """ Initialize the object.
 
-        Parameters:
+        Parameters
+        ----------
         loglikelihood: function
         ndim: int
             Dimensionality
@@ -69,6 +139,9 @@ class LogLikelihood:
             if True the function evaluations will be saved in the hdf5 file
         history_filename: string
             The filename where the history will go
+        blob: boolean
+            if True we expect the logl output to be a tuple of logl value and
+            a blob, otherwise it'll be logl value only
         """
         self.loglikelihood = loglikelihood
         self.pool = pool
@@ -79,28 +152,39 @@ class LogLikelihood:
         self.history_filename = history_filename
         self.ndim = ndim
         self.failed_save = False
+        self.blob = blob
         if save:
             self.history_init()
 
     def map(self, pars):
-        """ Evaluate the likelihood f-n on the list of vectors
+        """
+        Evaluate the likelihood function on the list of vectors
         The pool is used if it was provided when the object was created
+
+        Returns
+        -------
+        ret: The list of LoglOutput objects
         """
         if self.pool is None:
-            ret = np.array(list(map(self.loglikelihood, pars)))
+            ret = list([
+                LoglOutput(_, self.blob) for _ in map(self.loglikelihood, pars)
+            ])
         else:
-            ret = np.array(self.pool.map(self.loglikelihood, pars))
+            ret = [
+                LoglOutput(_, self.blob)
+                for _ in self.pool.map(self.loglikelihood, pars)
+            ]
         if self.save:
-            self.history_append(ret, pars)
+            self.history_append([_.val for _ in ret], pars)
         return ret
 
     def __call__(self, x):
         """
         Evaluate the likelihood f-n once
         """
-        ret = self.loglikelihood(x)
+        ret = LoglOutput(self.loglikelihood(x), self.blob)
         if self.save:
-            self.history_append([ret], [x])
+            self.history_append([ret.val], [x])
         return ret
 
     def history_append(self, logls, pars):
@@ -163,7 +247,7 @@ class LogLikelihood:
 
 class RunRecord:
     """
-    This is the class that saves the results of the nested
+    This is the dictionary like class that saves the results of the nested
     run so it is basically a collection of various lists of
     quantities
     """
@@ -189,7 +273,8 @@ class RunRecord:
             'it',  # iteration the live (now dead) point was proposed
             'n',  # number of live points interior to dead point
             'bounditer',  # active bound at a specific iteration
-            'scale'  # scale factor at each iteration
+            'scale',  # scale factor at each iteration
+            'blob'  # blobs output by the log-likelihood
         ]
         if dynamic:
             keys.extend([
@@ -197,8 +282,8 @@ class RunRecord:
                 # these are special since their length
                 # is == the number of batches
                 'batch_nlive',  # number of live points added in batch
-                'batch_bounds'
-            ])  # loglikelihood bounds used in batch
+                'batch_bounds'  # loglikelihood bounds used in batch
+            ])
         for k in keys:
             D[k] = []
         self.D = D
@@ -211,6 +296,487 @@ class RunRecord:
         for k in newD.keys():
             self.D[k].append(newD[k])
 
+    def __getitem__(self, k):
+        return self.D[k]
+
+    def __setitem__(self, k, v):
+        self.D[k] = v
+
+    def keys(self):
+        return self.D.keys()
+
+
+class DelayTimer:
+    """ Utility class that allows us to detect a certain
+    time has passed"""
+
+    def __init__(self, dt):
+        """ Initialise the time with delay of dt seconds
+
+        Parameters
+        ----------
+
+        dt: float
+            The number of seconds in the timer
+        """
+        self.dt = dt
+        self.last_time = time.time()
+
+    def is_time(self):
+        """
+        Returns true if more than self.dt seconds has passed
+        since the initialization or last call of successful is_time()
+
+        Returns
+        -------
+        ret: bool
+             True if specified amout of time has passed since the
+             initialization or last successful is_time() call
+        """
+        curt = time.time()
+        if curt - self.last_time > self.dt:
+            self.last_time = curt
+            return True
+        return False
+
+
+PrintFnArgs = namedtuple('PrintFnArgs',
+                         ['niter', 'short_str', 'mid_str', 'long_str'])
+
+
+def print_fn(results,
+             niter,
+             ncall,
+             add_live_it=None,
+             dlogz=None,
+             stop_val=None,
+             nbatch=None,
+             logl_min=-np.inf,
+             logl_max=np.inf,
+             pbar=None):
+    """
+    The default function used to print out results in real time.
+
+    Parameters
+    ----------
+
+    results : tuple
+        Collection of variables output from the current state of the sampler.
+        Currently includes:
+        (1) particle index,
+        (2) unit cube position,
+        (3) parameter position,
+        (4) ln(likelihood),
+        (5) ln(volume),
+        (6) ln(weight),
+        (7) ln(evidence),
+        (8) Var[ln(evidence)],
+        (9) information,
+        (10) number of (current) function calls,
+        (11) iteration when the point was originally proposed,
+        (12) index of the bounding object originally proposed from,
+        (13) index of the bounding object active at a given iteration,
+        (14) cumulative efficiency, and
+        (15) estimated remaining ln(evidence).
+
+    niter : int
+        The current iteration of the sampler.
+
+    ncall : int
+        The total number of function calls at the current iteration.
+
+    add_live_it : int, optional
+        If the last set of live points are being added explicitly, this
+        quantity tracks the sorted index of the current live point being added.
+
+    dlogz : float, optional
+        The evidence stopping criterion. If not provided, the provided
+        stopping value will be used instead.
+
+    stop_val : float, optional
+        The current stopping criterion (for dynamic nested sampling). Used if
+        the `dlogz` value is not specified.
+
+    nbatch : int, optional
+        The current batch (for dynamic nested sampling).
+
+    logl_min : float, optional
+        The minimum log-likelihood used when starting sampling. Default is
+        `-np.inf`.
+
+    logl_max : float, optional
+        The maximum log-likelihood used when stopping sampling. Default is
+        `np.inf`.
+
+    """
+    if pbar is None:
+        print_fn_fallback(results,
+                          niter,
+                          ncall,
+                          add_live_it=add_live_it,
+                          dlogz=dlogz,
+                          stop_val=stop_val,
+                          nbatch=nbatch,
+                          logl_min=logl_min,
+                          logl_max=logl_max)
+    else:
+        print_fn_tqdm(pbar,
+                      results,
+                      niter,
+                      ncall,
+                      add_live_it=add_live_it,
+                      dlogz=dlogz,
+                      stop_val=stop_val,
+                      nbatch=nbatch,
+                      logl_min=logl_min,
+                      logl_max=logl_max)
+
+
+def get_print_fn_args(results,
+                      niter,
+                      ncall,
+                      add_live_it=None,
+                      dlogz=None,
+                      stop_val=None,
+                      nbatch=None,
+                      logl_min=-np.inf,
+                      logl_max=np.inf):
+    # Extract results at the current iteration.
+    loglstar = results.loglstar
+    logz = results.logz
+    logzvar = results.logzvar
+    delta_logz = results.delta_logz
+    bounditer = results.bounditer
+    nc = results.nc
+    eff = results.eff
+
+    # Adjusting outputs for printing.
+    if delta_logz > 1e6:
+        delta_logz = np.inf
+    if logzvar >= 0. and logzvar <= 1e6:
+        logzerr = np.sqrt(logzvar)
+    else:
+        logzerr = np.nan
+    if logz <= -1e6:
+        logz = -np.inf
+    if loglstar <= -1e6:
+        loglstar = -np.inf
+
+    # Constructing output.
+    long_str = []
+    # long_str.append("iter: {:d}".format(niter))
+    if add_live_it is not None:
+        long_str.append("+{:d}".format(add_live_it))
+    short_str = list(long_str)
+    if nbatch is not None:
+        long_str.append("batch: {:d}".format(nbatch))
+    long_str.append("bound: {:d}".format(bounditer))
+    long_str.append("nc: {:d}".format(nc))
+    long_str.append("ncall: {:d}".format(ncall))
+    long_str.append("eff(%): {:6.3f}".format(eff))
+    short_str.append(long_str[-1])
+    long_str.append("loglstar: {:6.3f} < {:6.3f} < {:6.3f}".format(
+        logl_min, loglstar, logl_max))
+    short_str.append("logl*: {:6.1f}<{:6.1f}<{:6.1f}".format(
+        logl_min, loglstar, logl_max))
+    long_str.append("logz: {:6.3f} +/- {:6.3f}".format(logz, logzerr))
+    short_str.append("logz: {:6.1f}+/-{:.1f}".format(logz, logzerr))
+    mid_str = list(short_str)
+    if dlogz is not None:
+        long_str.append("dlogz: {:6.3f} > {:6.3f}".format(delta_logz, dlogz))
+        mid_str.append("dlogz: {:6.1f}>{:6.1f}".format(delta_logz, dlogz))
+    else:
+        long_str.append("stop: {:6.3f}".format(stop_val))
+        mid_str.append("stop: {:6.3f}".format(stop_val))
+
+    return PrintFnArgs(niter=niter,
+                       short_str=short_str,
+                       mid_str=mid_str,
+                       long_str=long_str)
+
+
+def print_fn_tqdm(pbar,
+                  results,
+                  niter,
+                  ncall,
+                  add_live_it=None,
+                  dlogz=None,
+                  stop_val=None,
+                  nbatch=None,
+                  logl_min=-np.inf,
+                  logl_max=np.inf):
+    """
+    This is a function that does the status printing using tqdm module
+    """
+    fn_args = get_print_fn_args(results,
+                                niter,
+                                ncall,
+                                add_live_it=add_live_it,
+                                dlogz=dlogz,
+                                stop_val=stop_val,
+                                nbatch=nbatch,
+                                logl_min=logl_min,
+                                logl_max=logl_max)
+
+    pbar.set_postfix_str(" | ".join(fn_args.long_str), refresh=False)
+    pbar.update(fn_args.niter - pbar.n)
+
+
+def print_fn_fallback(results,
+                      niter,
+                      ncall,
+                      add_live_it=None,
+                      dlogz=None,
+                      stop_val=None,
+                      nbatch=None,
+                      logl_min=-np.inf,
+                      logl_max=np.inf):
+    """
+    This is a function that does the status printing using just
+    standard printing into the console
+    """
+    fn_args = get_print_fn_args(results,
+                                niter,
+                                ncall,
+                                add_live_it=add_live_it,
+                                dlogz=dlogz,
+                                stop_val=stop_val,
+                                nbatch=nbatch,
+                                logl_min=logl_min,
+                                logl_max=logl_max)
+    niter, short_str, mid_str, long_str = (fn_args.niter, fn_args.short_str,
+                                           fn_args.mid_str, fn_args.long_str)
+
+    long_str = ["iter: {:d}".format(niter)] + long_str
+
+    # Printing.
+    long_str = ' | '.join(long_str)
+    mid_str = ' | '.join(mid_str)
+    short_str = '|'.join(short_str)
+    if sys.stderr.isatty() and hasattr(shutil, 'get_terminal_size'):
+        columns = shutil.get_terminal_size(fallback=(80, 25))[0]
+    else:
+        columns = 200
+    if columns > len(long_str):
+        sys.stderr.write("\r" + long_str + ' ' * (columns - len(long_str) - 2))
+    elif columns > len(mid_str):
+        sys.stderr.write("\r" + mid_str + ' ' * (columns - len(mid_str) - 2))
+    else:
+        sys.stderr.write("\r" + short_str + ' ' *
+                         (columns - len(short_str) - 2))
+    sys.stderr.flush()
+
+
+# List of results attributes as
+# Name, type, description, shape (if array)
+_RESULTS_STRUCTURE = [
+    ('logl', 'array[float]', 'Log likelihood', 'niter'),
+    ('samples_it', 'array[int]',
+     "the sampling iteration when the sample was proposed "
+     "(e.g., iteration 570)", 'niter'),
+    ('samples_id', 'array[int]',
+     'The unique ID of the sample XXX (within nlive points)', None),
+    ('samples_n', 'array[int]',
+     'The number of live points at the point when the sample was proposed',
+     'niter'),
+    ('samples_u', 'array[float]', '''The coordinates of live points in the
+    unit cube coordinate system''', 'niter,ndim'),
+    ('samples_v', 'array[float]', '''The coordinates of live points''',
+     'niter,ndim'),
+    ('samples', 'array',
+     '''the location (in original coordinates). Identical to samples_v''',
+     'niter,ndim'), ('niter', 'int', 'number of iterations', None),
+    ('ncall', 'int', 'Total number likelihood calls', None),
+    ('logz', 'array', 'Array of cumulative log(Z) integrals', 'niter'),
+    ('logzerr', 'array', 'Array of uncertainty of log(Z)', 'niter'),
+    ('logwt', 'array', 'Array of log-posterior weights', 'niter'),
+    ('eff', 'float', 'Sampling efficiency', None),
+    ('nlive', 'int', 'Number of live points for a static run', None),
+    ('logvol', 'array[float]', 'Logvolumes of dead points', 'niter'),
+    ('information', 'array[float]', 'Information Integral H', 'niter'),
+    ('bound', 'array[object]',
+     "the set of bounding objects used to condition proposals for the "
+     "base run", 'nbound'),
+    ('bound_iter', 'array[int]',
+     "index of the bound being used for an iteration that generated the point",
+     'niter'),
+    ('samples_bound', 'array[int]',
+     "The index of the bound that the corresponding sample was drawn from",
+     'niter'),
+    ('samples_batch', 'array[int]',
+     "Tracks the batch during which the samples were proposed", 'niter'),
+    ('batch_bounds', 'array[tuple]',
+     "The log-likelihood bounds used to run a batch.", 'nbatch'),
+    ('batch_nlive', 'array[int]',
+     "The number of live points used for  given batch", 'nbatch'),
+    ('scale', 'array[float]', "Scalar scale applied for proposals", 'niter'),
+    ('blob', 'array[]',
+     'The auxiliary blobs computed by the log-likelihood function', 'niter')
+]
+
+
+class Results:
+    """
+    Contains the full output of a run along with a set of helper
+    functions for summarizing the output.
+    The object is meant to be unchangeable record of the static or
+    dynamic nested run.
+
+    Results attributes (name, type, description, array size):
+    """
+
+    _ALLOWED = set([_[0] for _ in _RESULTS_STRUCTURE])
+
+    def __init__(self, key_values):
+        """
+        Initialize the results using the list of key value pairs
+        or a dictionary
+        Results([('logl', [1, 2, 3]), ('samples_it',[1,2,3])])
+        Results(dict(logl=[1, 2, 3], samples_it=[1,2,3]))
+        """
+        self._keys = []
+        self._initialized = False
+        if isinstance(key_values, dict):
+            key_values_list = key_values.items()
+        else:
+            key_values_list = key_values
+        for k, v in key_values_list:
+            assert k not in self._keys  # ensure no duplicates
+            assert k in Results._ALLOWED, k
+            self._keys.append(k)
+            setattr(self, k, copy.copy(v))
+        required_keys = ['samples_u', 'samples_id', 'logl', 'samples']
+        # TODO I need to add here logz, logzerr
+        # but that requires ensuring that merge_runs always computes logz
+        for k in required_keys:
+            if k not in self._keys:
+                raise ValueError('Key %s must be provided' % k)
+        if 'nlive' in self._keys:
+            self._dynamic = False
+        elif 'samples_n' in self._keys:
+            self._dynamic = True
+        else:
+            raise ValueError(
+                'Trying to construct results object without nlive '
+                'or samples_n information')
+        self._initialized = True
+
+    def __copy__(self):
+        # this will be a deep copy
+        return Results(self.asdict().items())
+
+    def copy(self):
+        '''
+        return a copy of the object
+        all numpy arrays will be copied too
+        '''
+        return self.__copy__()
+
+    def __setattr__(self, name, value):
+        if name[0] != '_' and self._initialized:
+            raise RuntimeError("Cannot set attributes directly")
+        super().__setattr__(name, value)
+
+    def __getitem__(self, name):
+        if name in self._keys:
+            return getattr(self, name)
+        else:
+            raise KeyError(name)
+
+    def __repr__(self):
+        m = max(list(map(len, list(self._keys)))) + 1
+        return '\n'.join(
+            [k.rjust(m) + ': ' + repr(getattr(self, k)) for k in self._keys])
+
+    def __contains__(self, key):
+        return key in self._keys
+
+    def keys(self):
+        """ Return the list of attributes/keys stored in Results """
+        return self._keys
+
+    def items(self):
+        """
+Return the list of items in the results object as list of key,value pairs
+        """
+        return ((k, getattr(self, k)) for k in self._keys)
+
+    def asdict(self):
+        """
+        Return contents of the Results object as dictionary
+        """
+        # importantly here we copy attribute values
+        return dict((k, copy.copy(getattr(self, k))) for k in self._keys)
+
+    def isdynamic(self):
+        """ Return true if the results was constructed using dynamic
+        nested sampling run with (potentially) variable number of
+        live-points"""
+        return self._dynamic
+
+    def importance_weights(self):
+        """
+        Return the importance weights for the each sample.
+        """
+        logwt = self['logwt'] - self['logz'][-1]
+        wt = np.exp(logwt)
+        wt = wt / wt.sum()
+        return wt
+
+    def samples_equal(self, rstate=None):
+        """
+        Return the equally weighted samples
+        """
+        if rstate is None:
+            rstate = get_random_generator()
+        return resample_equal(self['samples'],
+                              self.importance_weights(),
+                              rstate=rstate)
+
+    def summary(self):
+        """Return a formatted string giving a quick summary
+        of the results."""
+
+        if self._dynamic:
+            res = ("niter: {:d}\n"
+                   "ncall: {:d}\n"
+                   "eff(%): {:6.3f}\n"
+                   "logz: {:6.3f} +/- {:6.3f}".format(self['niter'],
+                                                      np.sum(self['ncall']),
+                                                      self['eff'],
+                                                      self['logz'][-1],
+                                                      self['logzerr'][-1]))
+        else:
+            res = ("nlive: {:d}\n"
+                   "niter: {:d}\n"
+                   "ncall: {:d}\n"
+                   "eff(%): {:6.3f}\n"
+                   "logz: {:6.3f} +/- {:6.3f}".format(self['nlive'],
+                                                      self['niter'],
+                                                      np.sum(self['ncall']),
+                                                      self['eff'],
+                                                      self['logz'][-1],
+                                                      self['logzerr'][-1]))
+
+        print('Summary\n=======\n' + res)
+
+
+Results.__doc__ += '\n\n' + str('\n'.join(
+    ['| ' + str(_) for _ in _RESULTS_STRUCTURE])) + '\n'
+
+
+def results_substitute(results, kw_dict):
+    """ This is an utility method that takes a Result object and
+substituted certain keys in it. It returns a copy object!
+    """
+    new_list = []
+    for k, w in results.items():
+        if k not in kw_dict:
+            new_list.append((k, w))
+        else:
+            new_list.append((k, kw_dict[k]))
+    return Results(new_list)
+
 
 def get_enlarge_bootstrap(sample, enlarge, bootstrap):
     """
@@ -307,12 +873,14 @@ def get_neff_from_logwt(logwt):
     Compute the number of effective samples from an array of unnormalized
     log-weights. We use Kish Effective Sample Size (ESS)  formula.
 
-    Parameters:
+    Parameters
+    ----------
     logwt: numpy array
         Array of unnormalized weights
 
-    Returns:
-    neff: int
+    Returns
+    -------
+    int
         The effective number of samples
     """
 
@@ -331,7 +899,7 @@ def unitcheck(u, nonbounded=None):
 
     if nonbounded is None:
         # No periodic boundary conditions provided.
-        return np.min(u) > 0 and np.max(u) < 1
+        return u.min() > 0 and u.max() < 1
     else:
         # Alternating periodic and non-periodic boundary conditions.
         unb = u[nonbounded]
@@ -726,7 +1294,6 @@ def compute_integrals(logl=None, logvol=None, reweight=None):
     # = LV_{i+1} - (LV_{i+1} -LV_i) + log(1-exp(LV_{i+1}-LV{i}))
     dlogvol = np.diff(logvol, prepend=0)
     logdvol = logvol - dlogvol + np.log1p(-np.exp(dlogvol))
-
     # logdvol is log(delta(volumes)) i.e. log (X_i-X_{i-1})
     logdvol2 = logdvol + math.log(0.5)
     # These are log(1/2(X_(i+1)-X_i))
@@ -934,6 +1501,7 @@ def resample_run(res, rstate=None, return_idx=False):
     new_res_dict = dict(niter=len(res.ncall[samp_idx]),
                         ncall=res.ncall[samp_idx],
                         eff=eff,
+                        blob=res.blob[samp_idx],
                         samples=res.samples[samp_idx],
                         samples_id=res.samples_id[samp_idx],
                         samples_it=res.samples_it[samp_idx],
@@ -1079,6 +1647,7 @@ def unravel_run(res, print_progress=True):
                      samples_id=res.samples_id[strand],
                      samples_it=res.samples_it[strand],
                      samples_u=res.samples_u[strand],
+                     blob=res.blob[strand],
                      logwt=saved_logwt,
                      logl=logl,
                      logvol=logvol,
@@ -1098,8 +1667,7 @@ def unravel_run(res, print_progress=True):
 
         # Print progress.
         if print_progress:
-            sys.stderr.write('\rStrand: {0}/{1}     '.format(
-                counter + 1, nstrands))
+            sys.stderr.write(f'\rStrand: {counter+1}/{nstrands}     ')
 
     return new_res
 
@@ -1164,8 +1732,7 @@ def merge_runs(res_list, print_progress=True):
                 counter += 1
                 # Print progress.
                 if print_progress:
-                    sys.stderr.write('\rMerge: {0}/{1}     '.format(
-                        counter, ntot))
+                    sys.stderr.write(f'\rMerge: {counter}/{ntot}     ')
             # Overwrite baseline set of results with merged results.
             rlist_base = copy.copy(rlist_new)
 
@@ -1184,7 +1751,7 @@ def merge_runs(res_list, print_progress=True):
         counter += 1
         # Print progress.
         if print_progress:
-            sys.stderr.write('\rMerge: {0}/{1}     '.format(counter, ntot))
+            sys.stderr.write(f'\rMerge: {counter}/{ntot}     ')
 
     res = check_result_static(res)
 
@@ -1277,11 +1844,10 @@ def kld_error(res,
         new_res, samp_idx = resample_run(res, rstate=rstate, return_idx=True)
         logp2 = logp2[samp_idx]  # re-order our original results to match
     else:
-        raise ValueError(
-            "Input `'error'` option '{0}' is not valid.".format(error))
+        raise ValueError("Input `'error'` option '{error}' is not valid.")
 
     # Define our new importance weights.
-    logp1 = new_res.logwt - new_res.logz[-1]
+    logp1 = new_res['logwt'] - new_res['logz'][-1]
 
     # Compute the KL divergence.
     kld = np.cumsum(np.exp(logp1) * (logp1 - logp2))
@@ -1325,7 +1891,8 @@ def _merge_two(res1, res2, compute_aux=False):
                      v=res1.samples,
                      logl=res1.logl,
                      nc=res1.ncall,
-                     it=res1.samples_it)
+                     it=res1.samples_it,
+                     blob=res1.blob)
     nbase = len(base_info['id'])
 
     # Number of live points throughout the run.
@@ -1358,7 +1925,8 @@ def _merge_two(res1, res2, compute_aux=False):
                     v=res2.samples,
                     logl=res2.logl,
                     nc=res2.ncall,
-                    it=res2.samples_it)
+                    it=res2.samples_it,
+                    blob=res2.blob)
     nnew = len(new_info['id'])
 
     # Number of live points throughout the run.
@@ -1398,7 +1966,8 @@ def _merge_two(res1, res2, compute_aux=False):
                          nc=[],
                          it=[],
                          n=[],
-                         batch=[])
+                         batch=[],
+                         blob=[])
 
     # Check if batch info is the same and modify counters accordingly.
     if np.all(base_info['bounds'] == new_info['bounds']):
@@ -1447,7 +2016,7 @@ def _merge_two(res1, res2, compute_aux=False):
             idx_new += 1
             combined_info['batch'].append(from_run['batch'][add_idx] + boffset)
 
-        for curk in ['id', 'u', 'v', 'logl', 'nc', 'it']:
+        for curk in ['id', 'u', 'v', 'logl', 'nc', 'it', 'blob']:
             combined_info[curk].append(from_run[curk][add_idx])
 
         # Save the number of live points and expected ln(volume).
@@ -1480,7 +2049,8 @@ def _merge_two(res1, res2, compute_aux=False):
              samples=np.asarray(combined_info['v']),
              logl=np.asarray(combined_info['logl']),
              logvol=np.asarray(combined_info['logvol']),
-             batch_bounds=np.asarray(bounds))
+             batch_bounds=np.asarray(bounds),
+             blob=np.asarray(combined_info['blob']))
 
     for curk in ['id', 'it', 'n', 'u', 'batch']:
         r['samples_' + curk] = np.asarray(combined_info[curk])
@@ -1588,28 +2158,27 @@ def old_stopping_function(results,
     pfrac = args.get('pfrac', 1.0)
     if not 0. <= pfrac <= 1.:
         raise ValueError(
-            "The provided `pfrac` {0} is not between 0. and 1.".format(pfrac))
+            f"The provided `pfrac` {pfrac} is not between 0. and 1.")
     evid_thresh = args.get('evid_thresh', 0.1)
     if pfrac < 1. and evid_thresh < 0.:
-        raise ValueError("The provided `evid_thresh` {0} is not non-negative "
-                         "even though `1. - pfrac` is {1}.".format(
-                             evid_thresh, 1. - pfrac))
+        raise ValueError(
+            f"The provided `evid_thresh` {evid_thresh} is not non-negative "
+            f"even though `pfrac` is {pfrac}.")
     post_thresh = args.get('post_thresh', 0.02)
     if pfrac > 0. and post_thresh < 0.:
-        raise ValueError("The provided `post_thresh` {0} is not non-negative "
-                         "even though `pfrac` is {1}.".format(
-                             post_thresh, pfrac))
+        raise ValueError(
+            f"The provided `post_thresh` {post_thresh} is not non-negative "
+            f"even though `pfrac` is {pfrac}.")
     n_mc = args.get('n_mc', 128)
     if n_mc <= 1:
-        raise ValueError("The number of realizations {0} must be greater "
-                         "than 1.".format(n_mc))
+        raise ValueError(f"The number of realizations {n_mc} must be greater "
+                         "than 1.")
     if n_mc < 20:
         warnings.warn("Using a small number of realizations might result in "
                       "excessively noisy stopping value estimates.")
     error = args.get('error', 'jitter')
     if error not in {'jitter', 'resample'}:
-        raise ValueError(
-            "The chosen `'error'` option {0} is not valid.".format(error))
+        raise ValueError(f"The chosen `'error'` option {error} is not valid.")
     approx = args.get('approx', True)
 
     # Compute realizations of ln(evidence) and the KL divergence.
@@ -1637,3 +2206,80 @@ def old_stopping_function(results,
         return stop <= 1., (stop_post, stop_evid, stop)
     else:
         return stop <= 1.
+
+
+def restore_sampler(fname, pool=None):
+    """
+    Restore the dynamic sampler from a file.
+    It is assumed that the file was created using .save() method
+    of DynamicNestedSampler or as a result of checkpointing during
+    run_nested()
+
+    Parameters
+    ----------
+    fname: string
+        Filename of the save file.
+    pool: object(optional)
+        The multiprocessing pool-like object that supports map()
+        calls that will be used in the restored object.
+
+    Returns
+    -------
+    Static or dynamic nested sampling object
+
+    """
+    with open(fname, 'rb') as fp:
+        res = pickle_module.load(fp)
+    sampler = res['sampler']
+    save_ver = res['version']
+    dynesty_format_version = 1
+    file_format_version = res['format_version']
+    if file_format_version != dynesty_format_version:
+        raise RuntimeError('Incorrect format version')
+    if save_ver != DYNESTY_VERSION:
+        warnings.warn(
+            f'The dynesty version in the checkpoint file ({save_ver})'
+            f'does not match the current dynesty version'
+            '({DYNESTY_VERSION}). That is *NOT* guaranteed to work')
+    if pool is not None:
+        sampler.M = pool.map
+        sampler.pool = pool
+        sampler.loglikelihood.pool = pool
+    else:
+        sampler.loglikelihood.pool = None
+        sampler.pool = None
+        sampler.M = map
+    return sampler
+
+
+def save_sampler(sampler, fname):
+    """
+    Save the state of the dynamic sampler in a file
+
+    Parameters
+    ----------
+    sampler: object
+        Dynamic or Static nested sampler
+    fname: string
+        Filename of the save file.
+
+    """
+    format_version = 1
+    # this is an internal version of the format we are
+    # using. Increase this if incompatible changes are being made
+    D = {
+        'sampler': sampler,
+        'version': DYNESTY_VERSION,
+        'format_version': format_version
+    }
+    tmp_fname = fname + '.tmp'
+    try:
+        with open(tmp_fname, 'wb') as fp:
+            pickle_module.dump(D, fp)
+        os.rename(tmp_fname, fname)
+    except:  # noqa
+        try:
+            os.unlink(tmp_fname)
+        except:  # noqa
+            pass
+        raise
